pro fit_main,dtime=dtime,fname1,fname3,psfile=psfile,outfile=outfile,proffile=profilefile

;   parameters and variables
;     variable names of input data comprehensive for the program
;
;	wave	: wavelength vector [Angstoem]
;	sp	: source count vector [cts]
;	err	: vector with errors [cts]
;	bkgd	: vector with instrumental + source background [cts]
;		  the source background is added only for the run
;		  of this program;
;		  bkgd will be returned in the original form.
;	dtime	: exposure time
;	fname1	: name of data file (string)
;
; ----------------------------------------------
;     other variable names used in the program
;       ampout  : actual value of amplitude
;       app_s/_b: appertures of source/backgr image
;       bbb     : smoothed background in search interval
;       bg0     : vector with weighted original instrumental background [cts]
;       bweight : weighting factor for background from different apperturas
;       eramp   : error in amplitude
;       erlam   : error in lambda
;       ersig   : error in sigma
;       fname1  : name of datafile (will be printed into the plot)
;       fname3  : name of input file
;       good    : index vector of actual analyzed range
;       idlam   : switch for constant distance of lines (-> steplam=const)
;       ierrplot: switch for drawing error bars in plot
;       iinput  : =0 read from COMMON block instead of input file (for autofit)
;       ilab    : labels will be insterted into plot at adequate positions
;       ilamit  : switch for iteration of lambda
;       imeth   : switch for specifying minimization method
;       iord    : investigate higher order of same line
;                 done by multiplication of starting value(s) of lambda with iord
;       iplot   : =1 create plot of fit
;	ilog	: switch for logarithmic plot
;	spec_only : =1 plot only the spectrum; no fitting
;       ires    : =1 print results on screen =0 print results into file 'fit.dat'
;       isigit  : switch for for iteration in sigma
;       ispline : switch for fitting liklihood curves with cubic splines
;       lamend  : upper limits of lambda
;       lamstart: lower limits of lambda
;       linID   : label of line
;       llambda : actual value of best lambda
;       lsig    : lower value for sigma
;	sigunit : units of sigma: =1 -> km/s else Angstroem
;       lsigma  : actual value of best sigma
;	f	: array with weight factors for auxiliary conditions
;	v	  in the form sum(f_i*ampout_i)=v
;       nbin    : number of bins beeing combined (integral will be conserved)
;       ncount  : count vector in search interval
;       nlik    : switch for plotting liklihood curves (also number of bins)
;       nlines  : number of lines
;       num_sig : parameter for width of interval
;		  within which likelihood is determined (cf. fname3)
;       sbg     : estimated continuum background of source
;       sbgit   : iterate sbg
;	steplam	: array with linedistances which can be forced to be constant
;       tol     : tolerance parameter for Brents and Powells iteration
;       unit    : channel number for reading input file
;       usig    : upper value for sigma
;       x       : wavelength vector in search interval good
;       yc      : vector containing theoretical curve of lines
;       exp	: exposure time

tol=1.E-4

@fit.inc

; style definition, if not set in fit.inc
if n_elements(cs) eq 0 then cs=1.4          ;charsize general
if n_elements(lcs) eq 0 then lcs=1.1         ;charsize of labels
; spectrum
if n_elements(iline) eq 0 then iline=0         ;linestyle
if n_elements(ithick) eq 0 then ithick=1        ;thickness of line
; best fit
if n_elements(fline) eq 0 then fline=0
if n_elements(fthick) eq 0 then fthick=4
; all lines
if n_elements(aline) eq 0 then aline=1         ;negative value will switch off
if n_elements(athick) eq 0 then athick=1
; background
if n_elements(bline) eq 0 then bline=0
if n_elements(bthick) eq 0 then bthick=1
; instr. background
if n_elements(inline) eq 0 then inline=2
if n_elements(inthick) eq 0 then inthick=1


if n_elements(dtime) eq 0 then dtime=0.

nwave=n_elements(wave)

l_read=fname1 ne 'NULL' and nwave le 0
if n_elements(datafile) ne 0 then l_read=l_read or datafile ne fname1
if l_read then read_data,unit,fname1

nwave=n_elements(wave)
binsize=(wave(nwave-1)-wave(0))/(nwave-1)

datafile=fname1 ; store filenames into COMMON blocks (cf. fit.inc)
conffile=fname3

if conffile ne 'NULL' then input,unit,conffile

if iplot gt 0 and iformat gt 0 and n_elements(psfile) eq 0 then begin
   print,'No PS file name is provided; will use cora_IDL.ps'
   psfile = 'cora_IDL.ps'
endif

; determine line distances to be kept constant if idlam=1
if nlines gt 1 then steplam=fltarr(nlines-1)
FOR lline=0,nlines-2 DO steplam(lline)=llambda(lline+1)-llambda(lline)

; initialize linevectors
ampout=fltarr(nlines)
erlam=fltarr(nlines)
ersig=fltarr(nlines)
eramp=fltarr(nlines)

;      determine dataintervall to be investigated
lamran,count

;      define current environment
  yc       = fltarr(nlines,count)
  x        = wave(good)
  dx	   = fltarr(count); x-shift(x,1)
  for j=0,count-1 do begin
    if (j eq 0) then dx(j)=x(1)-x(0)
    if (j eq count-1) then dx(j)=x(count-1)-x(count-2)
    if (j gt 0 and j lt count-1) then dx(j) = 0.5*(x(j+1)-x(j-1))
  endfor

  dx(0)=dx(1)
  ncount   = sp(good)
  errcount = err(good)
  bbb      = bkgd(good)

;      initialize profile
if n_elements(profilefile) ne 0 then begin
 proffile=profilefile
 num_profile_init
endif else prcount=0

;  set start values
ampstart = fltarr(nlines)
FOR lline=0,nlines-1 DO ampstart(lline) = total(ncount)-total(bbb)
ampstart=ampstart/nlines
 ampzero  = fltarr(nlines)

;      determine source background by median function
sbgmed=median(ncount-bbb)
if sbgit gt 0 then sbgiterate,ncount-bbb,sbgmed,sbg,sbgchi
if sbg lt 0 then sbg=0
sbgA=sbg/binsize

;      smoothing of background
bg0=bkgd
if bgsmooth gt 0 then begin
   smoothbg,wave(good),bg0(good),bkgdsmooth,2
   bkgd(good)=bkgdsmooth
endif

;  allow for different appertures of source and background exposure
bweight=float(app_s/app_b)
;  add additional source background
bg0=bg0*bweight
bkgd(good)=bweight*bkgd(good)+sbg
;  this action will be undone at the end of the program!
bbb      = bkgd(good)
 likezero = likelihood(ampzero)


if spec_only eq 0 then begin
 llik=func_lik([llambda,lsigma])
 if min(ampout) gt 0. then ampstart=ampout
 slik0=func_lik([llambda,lsig])
 slik1=func_lik([llambda,usig])
 llik0=func_lik([lamstart,lsigma])
 llik1=func_lik([lamend,lsigma])
 llik=func_lik([llambda,lsigma])
endif

;    plot likelihood curve for each parameter
if nlik gt 0 then begin
    set_plot,'ps'
    if iportrait eq 1 then $&
       DEVICE,filename='like.ps',/PORTRAIT,XSIZE=11,YSIZE=22 $&
    else $&
       DEVICE,filename='like.ps',/LANDSCAPE
    print,'Opened device like.ps'
    P=[llambda,lsigma]
    if ilamit gt 0 then $&
      FOR lline=0,nlines-1 DO plotlik,P,0,lline,min(lamstart),max(lamend),likvec
    if isigit gt 0 then $&
      FOR lline=0,nlines-1 DO plotlik,P,1,lline,lsig(lline),usig(lline),likvec
    device,/close
    set_plot,'x'
endif

;iiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiii
;   do iteration ;;;;;; iteration  ;;;;;; iteration  ;;;;;; iteration
;iiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiii

;   the three methods initially provided are obsolete now.
;    If Amoeba is desired though, the parameter imeth can be
;    changed either here or in fit_procs (routine input)
;
if imeth eq 1 then begin
  if ires eq 1 then begin
   print,' -----------------------'
   print,'| Iteration with Amoeba |'
   print,' -----------------------'
  endif
   lamrange=(lamend-lamstart)/2.
   sigrange=(usig-lsig)/2.
   if isigit ne 0 then begin
      if ilamit ne 0 then begin
	 if idlam eq 1 then begin
	    xmin = AMOEBA(1.0e-5, P0=[llambda(0),lsigma],$&
                SCALE=[lamrange(0),sigrange],$&
                FUNCTION_NAME='func_clik', FUNCTION_VALUE=llik, NMAX=700)
            IF N_ELEMENTS(xmin) EQ 1 THEN MESSAGE, 'AMOEBA failed to converge'
	    lsigma=xmin(1:nlines)
	 endif else begin
	    xmin = AMOEBA(1.0e-5, P0=[llambda,lsigma],$&
                SCALE=[lamrange,sigrange],$&
                FUNCTION_NAME='func_lik', FUNCTION_VALUE=llik, NMAX=700)
	    IF N_ELEMENTS(xmin) EQ 1 THEN MESSAGE, 'AMOEBA failed to converge'
            llambda=xmin(0:nlines-1)
	    lsigma=xmin(nlines:2*nlines-1)
	 endelse
      endif else begin
         xmin = AMOEBA(1.0e-5, P0=lsigma,$&
                SCALE=sigrange,$&
                FUNCTION_NAME='func_siglik', FUNCTION_VALUE=llik, NMAX=700)
         IF N_ELEMENTS(xmin) EQ 1 THEN MESSAGE, 'AMOEBA failed to converge'
         lsigma=xmin
      endelse
   endif else begin
      if ilamit ne 0 then begin
         xmin = AMOEBA(1.0e-5, P0=llambda,$&
                SCALE=lamrange,$&
                FUNCTION_NAME='func_lamlik', FUNCTION_VALUE=llik, NMAX=700)
         IF N_ELEMENTS(xmin) EQ 1 THEN MESSAGE, 'AMOEBA failed to converge'
         llambda=xmin
      endif
   endelse
   llik=mean(llik)
endif
if imeth eq 0 then begin
  if ires eq 1 then begin
   print,' ----------------------'
   print,'| Iteration with Brent |'
   print,' ----------------------'
  endif
   nbmax=50
   if ilamit gt 0 then begin
      BRENT,nbmax,lamstart(0),llambda(0),lamend(nlines-1),tol,xmin,slik0,llik,slik1,idlam,i,'func_clamlik'
      llambda(0)=xmin
      if ires eq 1 then print,i,' iterations for lambda!'
   endif
   if isigit gt 0 then begin
      BRENT,nbmax,lsig(0),lsigma(0),usig(0),tol,xmin,llik0,llik,llik1,idlam,i,'func_siglik'
      lsigma=xmin
      if ires eq 1 then print,i,' iterations for sigma!'
   endif 
endif

if imeth eq 2 then begin
  if ires eq 1 then begin
   print,' -----------------------'
   print,'| Iteration with Powell |'
   print,' -----------------------'
  endif
   if isigit gt 0 then begin
     if ilamit gt 0 then begin
	if idlam eq 1 then begin
	  xi=fltarr(nlines+1,nlines+1)
	  FOR i=0,nlines DO xi(i,i)=1.0
	  P=[llambda(0),lsigma]
	  POWELL,P,xi,tol,llik,'func_clik',iter=i
	  lsigma=P(1:nlines)
	endif else begin
	  xi=fltarr(2*nlines,2*nlines)
	  FOR i=0,nlines-1 DO xi(i,i)=1.0
	  FOR i=nlines,2*nlines-1 DO xi(i,i)=1.0
	  P=[llambda,lsigma]
	  POWELL,P,xi,tol,llik,'func_lik',iter=i
	  llambda=P(0:nlines-1)
	  lsigma=P(nlines:2*nlines-1)
	endelse
	if ires eq 1 then print,i,'steps for lambda and sigma'
     endif
     if ilamit eq 0 then begin
	xi=fltarr(nlines,nlines)
	FOR i=0,nlines-1 DO xi(i,i)=1.0
	POWELL,lsigma,xi,tol,llik,'func_siglik',iter=i
	if ires eq 1 then print,i,'steps for sigma'
     endif
   endif else begin
     if ilamit gt 0 then begin
        xi=fltarr(nlines,nlines)
        FOR i=0,nlines-1 DO xi(i,i)=1.0
        POWELL,llambda,xi,tol,llik,'func_lamlik',iter=i
        if ires eq 1 then print,i,'steps for lambda'
     endif
   endelse
endif
;iiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiii
;   do iteration ;;;;;; iteration  ;;;;;; iteration  ;;;;;; iteration
;iiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiii

;  Determine errors
if spec_only eq 0 then begin
 EH=1.  ;returns error with 68.3% precision
 sigshift=.1*lsigma
 lamshift=.1*llambda
 Amperr, yc,ampout,EH,like0,like1,like2,eramp
 if isigit gt 0 then para_err,lsigma,llambda,1,sigshift,EH,ersig
 if ilamit gt 0 then para_err,llambda,lsigma,0,lamshift,EH,erlam
endif

;  exclude infite amplitudes; rather take a zero...
sel=where(finite(ampout) eq 0,ctest)
if ctest gt 0 then ampout(sel)=0.

if spec_only eq 0 then begin
 print,'RESULT:'
 print,'======='
 lsout=strarr(nlines+1)
 lsout=[cut(llambda,7),' '+STRING("305B)]
 print,'lambda =',lsout
 lsout=[cut(erlam,7),' '+STRING("305B)]
 print,STRING("261B),'   ',lsout
 lsout=[cut(lsigma,7),' '+STRING("305B)]
 print,'sigma =',lsout
 lsout=[cut(ersig,7),' '+STRING("305B)]
 print,STRING("261B),'   ',lsout
 lsout=[cut(ampout,7),' cts']
 print,'A =',lsout
 lsout=[cut(eramp,7),' cts']
 print,STRING("261B),'   ',lsout
 print,'lik =',cut(llik,7)
 print,'sbg =',cut(sbgA,8),' cts/A'
 print,'========================'
endif

;  produce nice plot of fit
;  Opening plot device
;          the plot device must be closed after
;          applying the program
;          this way individual plot elements
;          can be added with e.g. xyouts,oplot etc.
if iformat gt 0 and iplot gt 0 then begin
   set_plot,'ps'
     if iportrait eq 1 then $&
        DEVICE,filename=psfile,/ENCAPSUL,YSIZE=11,XSIZE=22,/color $&
     else $&
        DEVICE,filename=psfile,/encapsul,/color
   print,'Opened device ',psfile
endif

if ilog eq 1 then plotfit,fname1,dtime,fname3,/log $
   else plotfit,fname1,dtime,fname3

;   print results into file
if ires gt 0 and n_elements(outfile) eq 0 then begin
   print,'No outfile name is provided; will use fit.dat'
   outfile = 'fit.dat'
endif

if ires gt 0 then begin
 openw,1,outfile,/append
 printf,1,'#   lambda[A]         Sigma[A]           A[cts]       sbg[cts/A]'
 FOR i=0,nlines-1 DO BEGIN
     printf,1,format='(F10.3,x,F8.5,2x,F7.4,x,F7.5,2x,F9.4,x,F9.4,2x,F9.2)',$
     llambda(i),erlam(i),lsigma(i),ersig(i),ampout(i),eramp(i),sbgA
 ENDFOR
 close,1
endif

;      restore original background
bg0=bg0/bweight
bkgd=bg0

return
end

